/*   BEM = Batch Equilibrium Model (calculates chemical equilibira)
     Copyright (C) 1996  J. Prenzel (jprenze@gwdg.de)
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <string.h>

#include "bem.h"

static int makdef(int sect, int nwords, char *wrd[]);
static int setcell(int sect, int nwords, char *wrd[], CLL **cl0);
static int makcinp(int sect, int nwords, char *wrd[], CLL  *cl0);
static int cisp1(int sect, int nwords, char *wrd[], CLL  *cl0);
static int makdpgas(int sect, int nwords, char *wrd[], CLL  *cl0);
static int makdelta(int step, double delta[]);
static int makvinp(int sect, int nwords, char *wrd[], CLL  *cl0);
static int makvout(int sect, int nwords, char *wrd[], CLL  *cl0);
static int makwpt(int sect, int nwords, char *wrd[], CLL  *cl0);
static int recint(char * rec, int sect);

/*..................................... tittra ...................*/
int tittra(char *tfile,  char *rfile){
  int i,rc,rcrdr,step=0,sect=1;
  char rec[MAXINREC];
  double delta[MAXELT+MAXIEX+MAXGAS];


  Deltat=1.;


  Tat.nam=tfile;
  Res.nam=rfile;
  Actfil=tfile;
  CALL(rc,openfiles('r'));
  CALL(rc,openfiles('t'));
  if(Op.s)CALL(rc,openfiles('s')); 
  if(Op.i)CALL(rc,openfiles('i')); 
  if(Op.q || Op.Q)CALL(rc,openfiles('q')); 
  hdgres();
  recfil(Tat);
  CALL(rc,openfiles('g'));
  CALL(rc,openfiles('l'));

  Ncells=0;

  rcrdr=0;
  while(rcrdr != 2){
    rcrdr=steprdr(Tat.fil,rec);

    if(rcrdr == 0){ 
      CALL(rc,recint(rec,sect));
    }

    if(rcrdr > 0){
      sect=2;
      Actfil=tfile;
      for(i=0;i<Ncells;i++){
	Cl=Dir[i]; 
	CALL(rc,makdelta(step,delta));
	Cl->step=step;
	Qind=0;
	if(Op.q || (Op.Q && step == 0)){
	  Qind=1;
	  getint(i,step);
	}
	rc=cttmng(delta);
	if(rc != 0){
	  bemerr(39,"","","",Cl->it2cnt);
	  return rc;
	}
	if(step == 0) CALL(rc,putprp());
	putout();
	if(Op.s) show01();
	if(Op.i) outint(i,step);
      }
      step++;
      CALL(rc,openfiles('f')); /* an empty temp. file for flux data */
    }

  }
  return 0;
}

/*-------------------------------------------- recint -----------------*/
/* interpretation of input-records (.t-file) */
int recint(char * rec, int sect){
  int nwords,rc;
  char rec2[MAXINREC],*wrd[MAXWORDS];
  CLL *cl0;

      strncpy(rec2,rec,MAXINREC);
      nwords=words(rec2,wrd);

      if(nwords==0 || *wrd[0]=='*' || *wrd[0]==0)return 0;;

      if(strcmp(wrd[0],"define") == 0){
	CALL(rc,makdef(sect,nwords,wrd));
	return 0;;
      }

      if(strcmp(wrd[0],"cell") == 0){
	CALL(rc,setcell(sect,nwords,wrd,&cl0));
	return 0;;
      }

      if(strcmp(wrd[0],"cinp") == 0){
	CALL(rc,makcinp(sect,nwords,wrd,cl0));
	return 0;;
      }

      if(strcmp(wrd[0],"cisp1") == 0){
	CALL(rc,cisp1(sect,nwords,wrd,cl0));
	return 0;;
      }

      if(strcmp(wrd[0],"dpgas") == 0){
	CALL(rc,makdpgas(sect,nwords,wrd,cl0));
	return 0;;
      }

       if(strcmp(wrd[0],"vinp") == 0){
	CALL(rc,makvinp(sect,nwords,wrd,cl0));
	return 0;;
      }

      if(strcmp(wrd[0],"vout") == 0){
	CALL(rc,makvout(sect,nwords,wrd,cl0));
	return 0;;
      }

      if(strcmp(wrd[0],"wptslt")==0){
	CALL(rc,makwpt(sect,nwords,wrd,cl0));
	return 0;;
      }

      if(strcmp(wrd[0],"put")==0){
	if(sect>1){
	  bemerr(32,wrd[0],wrd[1],wrd[2],0);
	  return -32;
	}
	fputs(rec,Putsfil); 
	return 0;;
      }

      bemerr(13,wrd[0],wrd[1],wrd[2],0);
      return -32;

}






/*-------------------------------------------- makdelta ---------------*/
int makdelta(int step, double delta[]){
  int ndlt,i,ii,rc,typok,stat;
  char key[MAXELTNAM];
  double a;

  ndlt=Cl->nelt+Cl->niex+Cl->ngas;
  for(ii=0;ii<ndlt;ii++){
    delta[ii]=0.;
  }

  if(step == 0) return 0;

  rewind(Flowfil);
  stat=0;
  while(1){
    rc=nxtkv(Flowfil,MAXELTNAM,key,&a);
    if(rc<0){
      bemerr(35,"","","",0);
      return -32;
    }
    if(rc==0){
      if(stat==0){
	if(strcmp(key,Cl->id) == 0){
	  stat=1;                     /* line applies */
	}
	else{
	  stat=-1;                    /* ignore line */
	}
	continue;
      }
      if(stat == 1){
	typok=0;
	if(strcmp(key,"H2O") == 0){
	  delta[0] += a;
	  typok=1;
	}
	for(i=2;i<Cl->nelt;i++){
	  if(strcmp(key,Cl->el[i]->name) == 0){
	    delta[i] += a;
	    typok=1;
	    break;
	  }
	}
	for(i=0;i<Cl->ngas;i++){
	  if(strcmp(key,Cl->gs[i]->name) == 0){
	    delta[Cl->nelt+Cl->niex+i] += a;
	    typok=1;
	    break;
	  }
	}
	if(typok==0){
	  bemerr(36,Cl->id,key,"",0);
	  return -32;
	}
      }
    }
    if(rc == 1){
      stat=0;                         /* line ended */
      continue;
    }
    if(rc==2)break;
  }
  return 0;
}


/*----------------------------------------------- makdef --------------*/
/* define a new reaction cell */
int makdef(int sect, int nwords, char *wrd[]){
  int rc;
  
  if(sect != 1){
    bemerr(32,wrd[0],wrd[1],wrd[2],0);
    return -32;
  }
  if(nwords<3){
    bemerr(31,"define","","",0);
    return -32;
  }
  if(strlen(wrd[1]) > MAXIDNAM){
    bemerr(33,wrd[1],"","",0);
    return -32;
  }
  if(strlen(wrd[2]) > MAXFILNAM){
    bemerr(33,wrd[2],"","",0);
    return -32;
  }

  CALL(rc,makecl(wrd[1]));
  Mod.nam=wrd[2];
  CALL(rc,openfiles('m'));
  recfil(Mod);
  Actfil=Mod.nam;
  CALL(rc,modinit());
  CALL(rc,init0());
  CALL(rc,slt0());
  CALL(rc,setx0()); 
  Dir[Ncells]=Cl;
  Ncells++;

return 0;
}

/*----------------------------------------------- setcell -------------*/
/* set the source reaction cell */
int setcell(int sect, int nwords, char *wrd[], CLL **cl0){
  int i,typok;
  
  if(sect != 2){
    bemerr(32,wrd[0],wrd[1],wrd[2],0);
    return -32;
  }
  if(nwords<2){
    bemerr(31,"cell","","",0);
    return -32;
  }
  if(strlen(wrd[1]) > MAXIDNAM){
    bemerr(33,wrd[1],"","",0);
    return -32;
  }

  typok=0;
  for(i=0;i<Ncells;i++){
    if(strcmp(Dir[i]->id,wrd[1]) == 0){
      *cl0=Dir[i];
      typok=1;
      break;
    }
  }
  if(typok == 0){
    bemerr(34,wrd[1],"","",0);
    return -32;
  }

  return 0;
}


/*----------------------------------------------- makcinp -------------*/
/* prepare changes due to cinp - line */
int makcinp(int sect, int nwords, char *wrd[], CLL  *cl0){
  int i,k,rc,typok;
  double a;
  
  if(sect != 2){
    bemerr(32,wrd[0],wrd[1],wrd[2],0);
    return -32;
  }
  if(nwords<3){
    bemerr(31,"cinp","","",0);
    return -32;
  }

  fprintf(Flowfil,"%s %d ",cl0->id,cl0->step);

  k=1;
  while(k<=nwords-1){
    if(*wrd[k] == '*')break;

    typok=0;
    for(i=2;i<cl0->nelt;i++){
      if(strcmp(cl0->el[i]->name,wrd[k]) == 0){
	if(k>=nwords){
	  bemerr(14,wrd[k],"","",0);
	  return -32;
	}
	k++;
	CALL(rc,dnum(wrd[k],wrd[k-1],&a));
	fprintf(Flowfil,"%s %g ",cl0->el[i]->name,a*1.e-6*Deltat);
	typok=1;
	break;
      }
    }
    if(typok == 0){
      bemerr(15,wrd[k],"cinp","",0);
      return -32;
    }

    k++;
  }
  fprintf(Flowfil,"\n");
  return 0;
}

/*----------------------------------------------- cisp1 -------------*/
/* prepare changes due to cisp1 - line */
/* input conditional to saturation of a given salt 

using: 
         cisp1 <el> <input> <slt> <sp>

         <el>   = Element
         <input> = chemical input in mol/system (same as with cinp-line)
         <slt>   = an expression that has the form of a salt formula
         <sp>    = the pIAP of <slt> from the preceding step is calculated.
                   if IAP/<sp> > 0.1 then only 10% of <input> is used
                   if IAP/<sp> > 1   then <input> is ignored
                   otherwise <input>is used as specified                  */
 
int cisp1(int sect, int nwords, char *wrd[], CLL  *cl0){
  int i,rc,typok;
  double a,piap,psp;
  ELT *el[MAXPARTS];
  int sc[MAXPARTS],n,z;

  if(sect != 2){
    bemerr(32,wrd[0],wrd[1],wrd[2],0);
    return -32;
  }
  if(nwords < 5){
    bemerr(31,"cisp1","","",0);
    return -32;
  }

  typok=0; /*.................................. get <el> & <input>   */
  for(i=2;i<cl0->nelt;i++){
    if(strcmp(cl0->el[i]->name,wrd[1]) == 0){
      CALL(rc,dnum(wrd[2],wrd[1],&a));
      typok=1;
      break;
    }
  }
  if(typok == 0){
    bemerr(15,wrd[1],"cisp1","",0);
    return -32;
  }
  
/*...................................... calculate IAP ................*/
  piap=0.;
  CALL(rc,cplfrm(wrd[3],el,sc,&n,&z));
  for(i=0;i<n;i++){
    piap += el[i]->pact * sc[i];
  }

  CALL(rc,dnum(wrd[4],"<sp>",&psp)); /*................. get <sp> .......*/

  if(piap > psp){ /* ................................... switch function */
    if(piap-psp <1) a *=0.1;
    fprintf(Flowfil,"%s %d ",cl0->id,cl0->step);
    fprintf(Flowfil,"%s %g ",wrd[1],a*1.e-6*Deltat);
    fprintf(Flowfil,"\n");
  }

  return 0;
}

/*----------------------------------------------- makdpgas -------------*/
/* prepare changes due to dpgas - line */
int makdpgas(int sect, int nwords, char *wrd[], CLL  *cl0){
  int i,k,rc,typok;
  double a;
  if(sect != 2){
    bemerr(32,wrd[0],wrd[1],wrd[2],0);
    return -32;
  }
  if(nwords<3){
    bemerr(31,"dpgas","","",0);
    return -32;
  }

  fprintf(Flowfil,"%s %d ",cl0->id,cl0->step);

  k=1;
  while(k<=nwords-1){
    if(*wrd[k] == '*')break;

    typok=0;
    for(i=0;i<cl0->ngas;i++){
      if(strcmp(cl0->gs[i]->name,wrd[k]) == 0){
	if(k>=nwords){
	  bemerr(14,wrd[k],"","",0);
	  return -32;
	}
	k++;
	CALL(rc,dnum(wrd[k],wrd[k-1],&a));
	fprintf(Flowfil,"%s %g ",cl0->gs[i]->name,a*1.e-3*Deltat);
	typok=1;
	break;
      }
    }
    if(typok == 0){
      bemerr(15,wrd[k],"dpgas","",0);
      return -32;
    }

    k++;
  }
  fprintf(Flowfil,"\n");
  return 0;
}

/*----------------------------------------------- makvinp -------------*/
/* prepare changes due to vinp - line */
int makvinp(int sect, int nwords, char *wrd[], CLL  *cl0){
  int i,k,rc,typok;
  double a,vol;
  
  if(sect != 2){
    bemerr(32,wrd[0],wrd[1],wrd[2],0);
    return -32;
  }
  if(nwords<2){
    bemerr(31,"vinp","","",0);
    return -32;
  }

  fprintf(Flowfil,"%s %d ",cl0->id,cl0->step);

  CALL(rc,dnum(wrd[1],wrd[0],&a));
  vol=a*1.e-3;
  fprintf(Flowfil,"H2O %g ",vol);

  k=2;
  while(k<=nwords-1){
    if(*wrd[k] == '*')break;

    typok=0;
    for(i=0;i<cl0->nelt;i++){
      if(strcmp(cl0->el[i]->name,wrd[k]) == 0){
	if(k>=nwords){
	  bemerr(14,wrd[k],"","",0);
	  return -32;
	}
	k++;
	CALL(rc,dnum(wrd[k],wrd[k-1],&a));
	fprintf(Flowfil,"%s %g ",cl0->el[i]->name,a*1.e-6*vol*Deltat);
	typok=1;
	break;
      }
    }
    if(typok == 0){
      bemerr(15,wrd[k],"vinp","",0);
      return -32;
    }

    k++;
  }
  fprintf(Flowfil,"\n");
  return 0;
}

/*----------------------------------------------- makvout -------------*/
/* prepare changes due to vout - line */
int makvout(int sect, int nwords, char *wrd[], CLL  *cl0){
  int i,k,rc,typok;
  double a,vol;
  char *toid;

  toid=NULL;
  if(sect != 2){
    bemerr(32,wrd[0],wrd[1],wrd[2],0);
    return -32;
  }
  if(nwords<2){
    bemerr(31,"vout","","",0);
    return -32;
  }

  fprintf(Flowfil,"%s %d ",cl0->id,cl0->step);

  CALL(rc,dnum(wrd[1],wrd[0],&a));
  vol=a*1.e-3;
  fprintf(Flowfil,"H2O %g ",-vol);

  k=2;
  while(k<=nwords-1){
    if(*wrd[k] == '*')break;
    typok=0;
    if(strcmp(wrd[k],"to") == 0){
      if(k>=nwords){
	bemerr(14,wrd[k],"","",0);
	return -32;
      }
      k++;
      for(i=0;i<Ncells;i++){
	if(strcmp(wrd[k],Dir[i]->id) == 0){
	  toid=wrd[k];
	  typok=1;
	  break;
	}
      }
    }

    if(typok == 0){
      bemerr(15,wrd[k],"vinp","",0);
      return -32;
    }

    k++;
  }

  for(i=2;i<cl0->nelt;i++){
    fprintf(Flowfil," %s %g "
            ,cl0->el[i]->name,-cl0->el[i]->cttl*vol*Deltat);
  }
  fprintf(Flowfil,"\n");

  if(toid != NULL){
    fprintf(Flowfil,"%s %d ",toid,cl0->step);
    fprintf(Flowfil,"H2O %g ",vol);
    for(i=2;i<cl0->nelt;i++){
      fprintf(Flowfil," %s %g ",cl0->el[i]->name,cl0->el[i]->cttl*vol*Deltat);
    }
    fprintf(Flowfil,"\n");
  }

  return 0;
}

/*----------------------------------------------- makwpt -------------*/
int makwpt(int sect, int nwords, char *wrd[], CLL  *cl0){
  int i,k,rc,typok;
  double a;
  
  if(sect != 2){
    bemerr(32,wrd[0],wrd[1],wrd[2],0);
    return -32;
  }
  if(nwords<3){
    bemerr(31,"vinp","","",0);
    return -32;
  }

  k=1;
  while(k<=nwords-1){
    if(*wrd[k] == '*')break;

    typok=0;
    for(i=0;i<cl0->nslt;i++){
      if(strcmp(cl0->sl[i]->name,wrd[k]) == 0){
	if(k>=nwords){
	  bemerr(14,wrd[k],"","",0);
	  return -32;
	}
	k++;
	CALL(rc,dnum(wrd[k],wrd[k-1],&a));
	cl0->sl[i]->wptfix=a*1.e-3;
	typok=1;
	break;
      }
    }
    if(typok == 0){
      bemerr(15,wrd[k],"wptslt","",0);
      return -32;
    }

    k++;
  }

  return 0;
}







